import unittest
import numpy as np
import struct
from lazperf import VLRDecompressor, VLRCompressor, RecordSchema, LazVLR


LAS_HEADER_SIZE = 227
VLR_HEADER_SIZE = 54
OFFSET_TO_LASZIP_VLR_DATA = LAS_HEADER_SIZE + VLR_HEADER_SIZE
LASZIP_VLR_DATA_SIZE = 52
OFFSET_TO_POINT_DATA = OFFSET_TO_LASZIP_VLR_DATA + LASZIP_VLR_DATA_SIZE
SIZEOF_CHUNK_TABLE_OFFSET = 8
POINT_COUNT = 1065

POINT_DTYPE = np.dtype(
    [
        ("X", "<i4"),
        ("Y", "<i4"),
        ("Z", "<i4"),
        ("intensity", "<u2"),
        ("bit_fields", "u1"),
        ("raw_classification", "u1"),
        ("scan_angle_rank", "i1"),
        ("user_data", "u1"),
        ("point_source_id", "<u2"),
        ("gps_time", "<f8"),
        ("red", "<u2"),
        ("green", "<u2"),
        ("blue", "<u2"),
    ]
)


class TestVLRDecompress(unittest.TestCase):
    def test_decompression(self):

        with open("test/simple_points_uncompressed.bin", mode="rb") as fin:
            points_ground_truth = np.frombuffer(fin.read(), dtype=POINT_DTYPE)

        with open("test/simple.laz", mode="rb") as fin:
            raw_data = fin.read()

        laszip_vlr_data = raw_data[
            OFFSET_TO_LASZIP_VLR_DATA : OFFSET_TO_LASZIP_VLR_DATA + LASZIP_VLR_DATA_SIZE
        ]

        raw_points = raw_data[OFFSET_TO_POINT_DATA + SIZEOF_CHUNK_TABLE_OFFSET :]

        laszip_vlr_data = np.frombuffer(laszip_vlr_data, dtype=np.uint8)
        compressed_points = np.frombuffer(raw_points, dtype=np.uint8)

        decompressor = VLRDecompressor(
            compressed_points, POINT_DTYPE.itemsize, laszip_vlr_data
        )
        points_decompressed = decompressor.decompress_points(POINT_COUNT)

        points_decompressed = np.frombuffer(points_decompressed, dtype=POINT_DTYPE)

        self.assertTrue(
            np.all(
                np.allclose(
                    points_ground_truth[dim_name], points_decompressed[dim_name]
                )
                for dim_name in POINT_DTYPE.names
            )
        )


class TestVLRCompress(unittest.TestCase):
    def test_compression(self):
        with open("test/simple_points_uncompressed.bin", mode="rb") as fin:
            points_to_compress = np.frombuffer(fin.read(), dtype=np.uint8)

        with open("test/simple.laz", mode="rb") as fin:
            ground_truth = fin.read()

        laszip_vlr_data = ground_truth[
            OFFSET_TO_LASZIP_VLR_DATA : OFFSET_TO_LASZIP_VLR_DATA + LASZIP_VLR_DATA_SIZE
        ]

        gt_vlr_data = np.frombuffer(laszip_vlr_data, dtype=np.uint8)

        rs = RecordSchema()
        rs.add_point()
        rs.add_gps_time()
        rs.add_rgb()

        vlr = LazVLR(rs)
        vlr_data = vlr.data()

        self.assertEqual(vlr.data_size(), LASZIP_VLR_DATA_SIZE)
        self.assertTrue(np.all(vlr_data == gt_vlr_data))

        offset_to_data = OFFSET_TO_LASZIP_VLR_DATA + vlr.data_size()
        self.assertEqual(offset_to_data, OFFSET_TO_POINT_DATA)

        compressor = VLRCompressor(rs, offset_to_data)
        points_compressed = compressor.compress(points_to_compress)

        chunk_table_offset = struct.unpack(
            "<Q", points_compressed[:SIZEOF_CHUNK_TABLE_OFFSET].tobytes()
        )[0]
        gt_chunk_table_offset = struct.unpack(
            "<Q",
            ground_truth[
                OFFSET_TO_POINT_DATA : OFFSET_TO_POINT_DATA + SIZEOF_CHUNK_TABLE_OFFSET
            ],
        )[0]
        self.assertEqual(chunk_table_offset, gt_chunk_table_offset)

        # This is the chunk table offset relative to the point data
        # without las header and vlrs
        relative_chunk_table_offset = chunk_table_offset - offset_to_data
        chunk_table = points_compressed[relative_chunk_table_offset:].tobytes()
        gt_chunk_table = ground_truth[gt_chunk_table_offset:]
        self.assertEqual(chunk_table, gt_chunk_table)

        points_compressed = points_compressed[SIZEOF_CHUNK_TABLE_OFFSET:].tobytes()
        gt_points_compressed = ground_truth[
            OFFSET_TO_POINT_DATA + SIZEOF_CHUNK_TABLE_OFFSET :
        ]
        self.assertEqual(points_compressed, gt_points_compressed)

    def test_raises(self):
        with open("test/simple_points_uncompressed.bin", mode="rb") as fin:
            point_buffer = fin.read()

        rs = RecordSchema()
        rs.add_gps_time()
        rs.add_rgb()

        vlr = LazVLR(rs)

        # Cut off point data to check if exception is properly raised
        points_to_compress = np.frombuffer(point_buffer[:-12], dtype=np.uint8)
        offset_to_data = OFFSET_TO_LASZIP_VLR_DATA + vlr.data_size()
        compressor = VLRCompressor(rs, offset_to_data)

        with self.assertRaises(ValueError):
            _ = compressor.compress(points_to_compress)


if __name__ == "__main__":
    unittest.main()



# distutils: language=c++
from libcpp.vector cimport vector
from libc.stdint cimport uint8_t, uint64_t

cdef extern from "PyLazperfTypes.hpp" namespace "pylazperf":
    enum LazPerfType "pylazperf::Type":
        Double
        Float
        Signed8
        Signed16
        Signed32
        Signed64
        Unsigned8
        Unsigned16
        Unsigned32
        Unsigned64



cdef extern from "PyLazperf.hpp" namespace "pylazperf":
    cdef cppclass Decompressor:
        Decompressor(const uint8_t *data, size_t dataLen) except +
        size_t decompress(char* buffer, size_t length)  except +
        size_t getPointSize()
        void add_dimension(LazPerfType t)



cdef extern from "PyLazperf.hpp" namespace "pylazperf":
    cdef cppclass Compressor:
        Compressor(vector[uint8_t]& arr) except +
        size_t compress(const char* buffer, size_t length)  except +
        size_t getPointSize()
        void add_dimension(LazPerfType t)
        void done()
        const vector[uint8_t]* data()
        void copy_data_to(uint8_t *dest) const



cdef extern from "laz-perf/factory.hpp" namespace "laszip::factory":
    cdef cppclass record_item:
        @staticmethod
        const record_item& point();
        @staticmethod
        const record_item& gpstime();
        @staticmethod
        const record_item& rgb();
        @staticmethod
        const record_item& eb(size_t count);

cdef extern from "laz-perf/factory.hpp" namespace "laszip::factory":
    cdef cppclass record_schema:
        record_schema()
        void push(const record_item&)


cdef extern from 'laz-perf/io.hpp' namespace "laszip::io":
    cdef cppclass laz_vlr:
        laz_vlr()
        void extract(char *data)
        size_t size() const
        @staticmethod
        laz_vlr from_schema(const record_schema)

cdef extern from "PyLazperf.hpp" namespace "pylazperf":
    cdef cppclass VlrDecompressor:
        VlrDecompressor(const unsigned char* compressed_data, size_t dataLength, size_t pointSize, const char *vlr) except +
        size_t decompress(char* buffer)  except +
        size_t getPointSize()


cdef extern from "PyLazperf.hpp" namespace "pylazperf":
    cdef cppclass VlrCompressor:
        VlrCompressor(record_schema, uint64_t offset) except +
        size_t compress(const char *inbuf) except +
        void done()
        const vector[uint8_t]* data()
        size_t getPointSize()
        size_t vlrDataSize() const
        void extractVlrData(char* out)
        void copy_data_to(uint8_t *dst) const


import struct
import unittest

import numpy as np
from lazperf import VLRDecompressor, RecordSchema, LazVLR, VLRCompressor

LAS_HEADER_SIZE = 227
LAZ_OFFSET_TO_POINTS = 1353
LAS_VLR_HEADER_SIZE = 54
LAS_POINT_SIZE = 61  # Extra-bytes included
LAS_POINT_COUNT = 1065

LASZIP_VLR_DATA_SIZE = 58
LASZIP_CHUNK_TABLE_SIZE = 8
OFFSET_TO_LASZIP_VLR_DATA = LAZ_OFFSET_TO_POINTS - LASZIP_VLR_DATA_SIZE

POINTS_DTYPE = np.dtype(
    [
        ("X", "<i4"),
        ("Y", "<i4"),
        ("Z", "<i4"),
        ("intensity", "<u2"),
        ("bit_fields", "u1"),
        ("raw_classification", "u1"),
        ("scan_angle_rank", "i1"),
        ("user_data", "u1"),
        ("point_source_id", "<u2"),
        ("gps_time", "<f8"),
        ("red", "<u2"),
        ("green", "<u2"),
        ("blue", "<u2"),
        ("Colors", "<u2", (3,)),
        ("Reserved", "u1", (7,)),
        ("Flags", "i1", (2,)),
        ("Intensity", "<u4"),
        ("Time", "<u8"),
    ]
)


class TestExtraBytesDecompression(unittest.TestCase):
    def test_decompression(self):
        with open("../test/raw-sets/extrabytes.laz", mode="rb") as f:
            raw_laz = f.read()

        # The file contains 2 vlrs in this order:
        # ExtraBytesVLR, LaszipVLR
        laszip_vlr_data = raw_laz[
            OFFSET_TO_LASZIP_VLR_DATA : OFFSET_TO_LASZIP_VLR_DATA + LASZIP_VLR_DATA_SIZE
        ]
        self.assertEqual(len(laszip_vlr_data), LASZIP_VLR_DATA_SIZE)

        raw_compressed_points = np.frombuffer(
            raw_laz[LAZ_OFFSET_TO_POINTS + LASZIP_CHUNK_TABLE_SIZE :], np.uint8
        )
        laszip_vlr_data = np.frombuffer(laszip_vlr_data, dtype=np.uint8)

        decompressor = VLRDecompressor(
            raw_compressed_points, POINTS_DTYPE.itemsize, laszip_vlr_data
        )
        decompressed_points = decompressor.decompress_points(LAS_POINT_COUNT)
        decompressed_points = np.frombuffer(decompressed_points, dtype=POINTS_DTYPE)

        with open("../test/raw-sets/extrabytes.las", mode="rb") as f:
            raw_las = f.read()

        offset_to_points = (
            LAZ_OFFSET_TO_POINTS - LAS_VLR_HEADER_SIZE - LASZIP_VLR_DATA_SIZE
        )
        expected_decompressed_points = np.frombuffer(
            raw_las[offset_to_points:], dtype=POINTS_DTYPE
        )

        self.assertTrue(
            np.all(
                np.allclose(
                    expected_decompressed_points[dim_name],
                    decompressed_points[dim_name],
                )
                for dim_name in POINTS_DTYPE.names
            )
        )


class TestExtraBytesCompression(unittest.TestCase):
    def test_compression(self):
        with open("../test/raw-sets/extrabytes.laz", mode="rb") as f:
            raw_laz = f.read()

        with open("../test/raw-sets/extrabytes.las", mode="rb") as f:
            raw_las = f.read()

        record_schema = RecordSchema()
        record_schema.add_point()
        record_schema.add_gps_time()
        record_schema.add_rgb()
        record_schema.add_extra_bytes(27)

        vlr = LazVLR(record_schema)
        vlr_data = vlr.data()
        expected_vlr_data = np.frombuffer(
            raw_laz[
                OFFSET_TO_LASZIP_VLR_DATA : OFFSET_TO_LASZIP_VLR_DATA
                + LASZIP_VLR_DATA_SIZE
            ],
            np.uint8,
        )

        self.assertEqual(LASZIP_VLR_DATA_SIZE, vlr.data_size())
        self.assertTrue(np.all(expected_vlr_data == vlr_data))

        compressor = VLRCompressor(record_schema, LAZ_OFFSET_TO_POINTS)
        points_to_compress = np.frombuffer(
            raw_las[
                LAZ_OFFSET_TO_POINTS - LASZIP_VLR_DATA_SIZE - LAS_VLR_HEADER_SIZE :
            ],
            np.uint8,
        )
        points_compressed = compressor.compress(points_to_compress)

        chunk_table_offset = struct.unpack(
            "<Q", points_compressed[:LASZIP_CHUNK_TABLE_SIZE].tobytes()
        )[0]
        expected_chunk_table_offset = struct.unpack(
            "<Q",
            raw_laz[
                LAZ_OFFSET_TO_POINTS : LAZ_OFFSET_TO_POINTS + LASZIP_CHUNK_TABLE_SIZE
            ],
        )[0]
        self.assertEqual(chunk_table_offset, expected_chunk_table_offset)

        relative_chunk_table_offset = chunk_table_offset - LAZ_OFFSET_TO_POINTS
        chunk_table = points_compressed[relative_chunk_table_offset:].tobytes()
        expected_chunk_table = raw_laz[expected_chunk_table_offset:]
        self.assertEqual(chunk_table, expected_chunk_table)

        expected_points_compressed = raw_laz[
            LAZ_OFFSET_TO_POINTS + LASZIP_CHUNK_TABLE_SIZE :
        ]
        points_compressed = points_compressed[LASZIP_CHUNK_TABLE_SIZE:].tobytes()
        self.assertEqual(expected_points_compressed, points_compressed)



#!/usr/bin/env python

# Stolen from Shapely's setup.py
#

import logging
import platform
import sys
from distutils.version import StrictVersion

from setuptools import setup
from setuptools.extension import Extension as DistutilsExtension
from setuptools.command.build_ext import build_ext as _build_ext

logging.basicConfig()
log = logging.getLogger(__file__)

DEBUG = True
USE_CYTHON = False
try:

    from Cython.Build import cythonize
    import cython

    if not StrictVersion(cython.__version__) < StrictVersion("0.24"):
        USE_CYTHON = True

except ImportError:
    pass

class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        try:
            __builtins__.__NUMPY_SETUP__ = False
        except AttributeError:
            __builtins__['__NUMPY_SETUP__'] = False

        import numpy
        self.include_dirs.append(numpy.get_include())
ext = ".pyx" if USE_CYTHON else ".cpp"

# python -W all setup.py ...
if "all" in sys.warnoptions:
    log.level = logging.DEBUG

# Get the version from the lazperf module
module_version = None
with open("lazperf/__init__.py", "r") as fp:
    for line in fp:
        if line.startswith("__version__"):
            module_version = StrictVersion(line.split("=")[1].strip().strip("\"'"))
            break

if not module_version:
    raise ValueError("Could not determine lazperf's version")

# Handle UTF-8 encoding of certain text files.
open_kwds = {}
if sys.version_info >= (3,):
    open_kwds["encoding"] = "utf-8"

with open("VERSION.txt", "w", **open_kwds) as fp:
    fp.write(str(module_version))

with open("README.rst", "r", **open_kwds) as fp:
    readme = fp.read()

with open("CHANGES.rst", "r", **open_kwds) as fp:
    changes = fp.read()

long_description = readme + "\n\n" + changes

include_dirs = ["."]

library_dirs = []
libraries = []
extra_link_args = []
extra_compile_args = []


# Here we assume everyone on windows uses the MSVC compiler
if platform.system() != "Windows":
    extra_compile_args = ["-std=c++11"]
    if DEBUG:
        extra_compile_args += ["-g", "-O0"]
else:
    include_dirs = ['..']



sources = [
    "lazperf/pylazperfapi" + ext,
    "lazperf/PyLazperf.cpp",
    "lazperf/PyVlrCompressor.cpp",
]
extensions = [
    DistutilsExtension(
        "lazperf.pylazperfapi",
        sources,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        language="c++",
    )
]

if USE_CYTHON and "clean" not in sys.argv:
    from Cython.Build import cythonize

    extensions = cythonize(extensions, language="c++")

setup_args = dict(
    name="lazperf",
    version=str(module_version),
    setup_requires=["numpy >=1.11"],
    description="Point cloud data compression",
    license="LGPL",
    keywords="point cloud compression",
    author="Howard Butler",
    author_email="howard@hobu.co",
    maintainer="Howard Butler",
    maintainer_email="howard@hobu.co",
    cmdclass={'build_ext':build_ext},
    url="https://github.com/hobu/laz-perf",
    long_description=long_description,
    test_suite="test",
    include_package_data=True,
    packages=["lazperf"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
)
setup(ext_modules=extensions, **setup_args)
